type token =
  | CALL_OF
  | RETURN_OF
  | CALLORRETURN_OF
  | IDENTIFIER of (string)
  | METAVAR of (string)
  | INT of (string)
  | AFF
  | LCURLY
  | RCURLY
  | LPAREN
  | RPAREN
  | LSQUARE
  | RSQUARE
  | LBRACELBRACE
  | RBRACERBRACE
  | RARROW
  | TRUE
  | FALSE
  | NOT
  | DOT
  | AMP
  | COLON
  | SEMI_COLON
  | COMMA
  | PIPE
  | CARET
  | QUESTION
  | COLUMNCOLUMN
  | EQ
  | LT
  | GT
  | LE
  | GE
  | NEQ
  | PLUS
  | MINUS
  | SLASH
  | STAR
  | PERCENT
  | OR
  | AND
  | OTHERWISE
  | EOF

open Parsing;;
let _ = parse_error;;
# 30 "src/plugins/aorai/yaparser.mly"
open Cil_types
open Logic_ptree
open Promelaast
open Bool3

type options =
  | Deterministic
  | Init of string list
  | Accept of string list
  | Observables of string list

let to_seq c =
  [{ condition = Some c;
     nested = [];
     min_rep = Some Data_for_aorai.cst_one;
     max_rep = Some Data_for_aorai.cst_one;
   }]

let is_no_repet (min,max) =
  let is_one c = Option.fold ~some:Data_for_aorai.is_cst_one ~none:false c in
  is_one min && is_one max

let observed_states      = Hashtbl.create 1
let prefetched_states    = Hashtbl.create 1

let fetch_and_create_state name =
  Hashtbl.remove prefetched_states name ;
  try
    Hashtbl.find observed_states name
  with
    Not_found ->
      let s = Data_for_aorai.new_state name in
      Hashtbl.add observed_states name s; s
;;

let prefetch_and_create_state name =
    if (Hashtbl.mem prefetched_states name) ||
      not (Hashtbl.mem observed_states name)
    then
      begin
        let s= fetch_and_create_state name in
        Hashtbl.add prefetched_states name name;
        s
      end
    else
      (fetch_and_create_state name)
;;

let set_init_state id =
  try
    let state = Hashtbl.find observed_states id in
    state.init <- True
  with Not_found ->
    Aorai_option.abort "no state '%s'" id

let set_accept_state id =
  try
    let state = Hashtbl.find observed_states id in
    state.acceptation <- True
  with Not_found ->
    Aorai_option.abort "no state '%s'" id

let add_metavariable map (name,typename) =
  let ty = match typename with
    | "int" -> TInt(IInt, [])
    | "char" -> TInt(IChar, [])
    | "long" -> TInt(ILong, [])
    | _ ->
      Aorai_option.abort "Unrecognized type %s for metavariable %s"
        typename name
  in
  if Datatype.String.Map.mem name map then
    Aorai_option.abort "The metavariable %s is declared twice" name;
  let vi =
    Cil.makeGlobalVar
      ~ghost:true (Data_for_aorai.get_fresh ("aorai_" ^ name)) ty
  in
  Datatype.String.Map.add name vi map

let check_state st =
  if st.acceptation=Undefined || st.init=Undefined then
   Aorai_option.abort
    "Error: the state '%s' is used but never defined." st.name

let interpret_option auto = function
  | Init states ->
    List.iter set_init_state states; auto
  | Accept states ->
    List.iter set_accept_state states; auto
  | Deterministic ->
    Aorai_option.Deterministic.set true; auto
  | Observables names ->
    let module Set = Datatype.String.Set in
    let new_set = Set.of_list names in
    let observables =
      match auto.observables with
      | None -> Some new_set
      | Some set -> Some (Set.union set new_set)
    in
    { auto with observables }

let build_automaton options metavariables trans =
  let htable_to_list table = Hashtbl.fold (fun _ st l -> st :: l) table [] in
  let states = htable_to_list observed_states
  and undefined_states = htable_to_list prefetched_states
  and metavariables =
    List.fold_left add_metavariable Datatype.String.Map.empty metavariables
  in
  let auto = { states; trans; metavariables; observables = None } in
  let auto = List.fold_left interpret_option auto options in
  List.iter check_state states;
  if not (List.exists (fun st -> st.init=True) states) then
    Aorai_option.abort "Automaton does not declare an initial state";
  if undefined_states <> [] then
    Aorai_option.abort "Error: the state(s) %a are used but never defined."
      (Pretty_utils.pp_list ~sep:"," Format.pp_print_string) undefined_states;
  auto

type pre_cond = Behavior of string | Pre of Promelaast.condition


# 171 "src/plugins/aorai/yaparser.ml"
let yytransl_const = [|
  257 (* CALL_OF *);
  258 (* RETURN_OF *);
  259 (* CALLORRETURN_OF *);
  263 (* AFF *);
  264 (* LCURLY *);
  265 (* RCURLY *);
  266 (* LPAREN *);
  267 (* RPAREN *);
  268 (* LSQUARE *);
  269 (* RSQUARE *);
  270 (* LBRACELBRACE *);
  271 (* RBRACERBRACE *);
  272 (* RARROW *);
  273 (* TRUE *);
  274 (* FALSE *);
  275 (* NOT *);
  276 (* DOT *);
  277 (* AMP *);
  278 (* COLON *);
  279 (* SEMI_COLON *);
  280 (* COMMA *);
  281 (* PIPE *);
  282 (* CARET *);
  283 (* QUESTION *);
  284 (* COLUMNCOLUMN *);
  285 (* EQ *);
  286 (* LT *);
  287 (* GT *);
  288 (* LE *);
  289 (* GE *);
  290 (* NEQ *);
  291 (* PLUS *);
  292 (* MINUS *);
  293 (* SLASH *);
  294 (* STAR *);
  295 (* PERCENT *);
  296 (* OR *);
  297 (* AND *);
  298 (* OTHERWISE *);
    0 (* EOF *);
    0|]

let yytransl_block = [|
  260 (* IDENTIFIER *);
  261 (* METAVAR *);
  262 (* INT *);
    0|]

let yylhs = "\255\255\
\001\000\002\000\002\000\005\000\006\000\006\000\007\000\007\000\
\003\000\003\000\008\000\008\000\009\000\004\000\004\000\010\000\
\011\000\011\000\012\000\012\000\012\000\015\000\015\000\016\000\
\016\000\017\000\017\000\017\000\017\000\018\000\018\000\019\000\
\019\000\013\000\013\000\021\000\021\000\021\000\021\000\021\000\
\021\000\021\000\021\000\020\000\020\000\023\000\023\000\024\000\
\024\000\024\000\024\000\024\000\024\000\024\000\024\000\025\000\
\025\000\025\000\025\000\025\000\025\000\022\000\022\000\022\000\
\026\000\026\000\026\000\026\000\027\000\027\000\027\000\027\000\
\028\000\028\000\028\000\029\000\029\000\029\000\029\000\030\000\
\030\000\030\000\030\000\030\000\014\000\014\000\031\000\000\000"

let yylen = "\002\000\
\004\000\002\000\001\000\004\000\000\000\002\000\003\000\001\000\
\000\000\001\000\002\000\001\000\004\000\002\000\001\000\004\000\
\003\000\001\000\006\000\004\000\003\000\001\000\003\000\000\000\
\001\000\003\000\006\000\005\000\004\000\002\000\003\000\000\000\
\003\000\001\000\002\000\000\000\001\000\001\000\001\000\005\000\
\003\000\004\000\004\000\003\000\001\000\003\000\001\000\004\000\
\004\000\004\000\001\000\001\000\002\000\003\000\001\000\003\000\
\003\000\003\000\003\000\003\000\003\000\003\000\003\000\001\000\
\003\000\003\000\003\000\001\000\001\000\003\000\003\000\003\000\
\001\000\002\000\001\000\003\000\003\000\004\000\001\000\002\000\
\005\000\001\000\003\000\001\000\000\000\002\000\004\000\002\000"

let yydefred = "\000\000\
\000\000\000\000\000\000\088\000\000\000\003\000\000\000\000\000\
\000\000\002\000\000\000\012\000\000\000\000\000\000\000\000\000\
\000\000\015\000\011\000\008\000\000\000\004\000\000\000\000\000\
\001\000\014\000\000\000\013\000\000\000\000\000\000\000\000\000\
\018\000\000\000\000\000\007\000\000\000\000\000\000\000\000\000\
\000\000\084\000\073\000\000\000\000\000\051\000\052\000\000\000\
\000\000\000\000\000\000\000\000\034\000\000\000\000\000\000\000\
\055\000\000\000\000\000\069\000\000\000\079\000\000\000\016\000\
\000\000\000\000\086\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\053\000\074\000\000\000\000\000\000\000\039\000\037\000\
\038\000\035\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\017\000\021\000\
\000\000\000\000\087\000\000\000\000\000\000\000\000\000\000\000\
\000\000\030\000\000\000\054\000\083\000\000\000\026\000\000\000\
\000\000\000\000\056\000\057\000\058\000\059\000\060\000\061\000\
\044\000\046\000\062\000\063\000\000\000\000\000\000\000\070\000\
\071\000\072\000\000\000\077\000\076\000\020\000\000\000\049\000\
\050\000\048\000\000\000\000\000\029\000\000\000\031\000\025\000\
\000\000\023\000\000\000\000\000\041\000\000\000\078\000\000\000\
\081\000\028\000\000\000\019\000\043\000\042\000\000\000\033\000\
\027\000\040\000"

let yydgoto = "\002\000\
\004\000\005\000\009\000\017\000\006\000\014\000\021\000\011\000\
\012\000\018\000\032\000\033\000\080\000\034\000\160\000\161\000\
\052\000\077\000\157\000\053\000\090\000\054\000\055\000\056\000\
\057\000\058\000\059\000\060\000\061\000\062\000\035\000"

let yysindex = "\024\000\
\006\255\000\000\046\255\000\000\012\255\000\000\050\255\054\255\
\089\255\000\000\092\255\000\000\105\255\093\255\116\255\103\255\
\001\000\000\000\000\000\000\000\117\255\000\000\119\255\028\255\
\000\000\000\000\139\255\000\000\133\255\063\255\150\255\071\255\
\000\000\147\255\150\255\000\000\108\255\154\255\171\255\176\255\
\016\255\000\000\000\000\085\255\063\255\000\000\000\000\085\255\
\173\255\014\255\179\255\100\255\000\000\222\255\151\255\155\255\
\000\000\095\255\053\255\000\000\138\255\000\000\196\255\000\000\
\028\255\210\255\000\000\209\255\108\255\198\255\218\255\225\255\
\241\255\010\255\085\255\243\255\214\255\246\255\173\000\226\255\
\247\255\000\000\000\000\138\255\150\255\255\254\000\000\000\000\
\000\000\000\000\108\255\108\255\108\255\108\255\108\255\108\255\
\085\255\085\255\108\255\108\255\108\255\108\255\108\255\108\255\
\108\255\108\255\108\255\002\000\003\000\018\000\000\000\000\000\
\013\000\015\000\000\000\017\000\019\000\020\000\057\255\021\000\
\014\000\000\000\063\255\000\000\000\000\063\255\000\000\022\000\
\108\255\029\255\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\053\255\053\255\053\255\000\000\
\000\000\000\000\026\000\000\000\000\000\000\000\035\000\000\000\
\000\000\000\000\085\255\029\000\000\000\023\000\000\000\000\000\
\046\000\000\000\030\000\031\000\000\000\101\255\000\000\044\000\
\000\000\000\000\023\000\000\000\000\000\000\000\052\000\000\000\
\000\000\000\000"

let yyrindex = "\000\000\
\000\000\000\000\000\000\000\000\058\000\000\000\040\000\000\000\
\000\000\000\000\060\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\042\000\000\000\000\000\050\000\
\000\000\000\000\000\000\000\000\000\000\000\000\050\000\000\000\
\000\000\000\000\050\000\000\000\000\000\000\000\000\000\000\000\
\161\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\207\255\000\000\000\000\174\255\113\255\
\000\000\111\000\235\255\000\000\169\255\000\000\000\000\000\000\
\050\000\000\000\000\000\136\255\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\087\255\
\000\000\000\000\000\000\202\255\050\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\145\000\000\000\
\000\000\000\000\056\000\000\000\000\000\104\255\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\012\000\045\000\078\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\145\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000\145\000\000\000\000\000\000\000\000\000\000\000\
\000\000\000\000"

let yygindex = "\000\000\
\000\000\000\000\000\000\000\000\065\001\000\000\000\000\000\000\
\060\001\055\001\000\000\008\001\058\001\231\255\213\255\220\000\
\000\000\000\000\132\255\220\255\000\000\219\255\250\000\046\001\
\000\000\000\000\238\255\192\255\045\001\000\000\000\000"

let yytablesize = 463
let yytable = "\070\000\
\025\000\081\000\068\000\042\000\043\000\063\000\079\000\078\000\
\069\000\067\000\038\000\039\000\040\000\041\000\042\000\043\000\
\008\000\068\000\042\000\044\000\119\000\045\000\129\000\069\000\
\001\000\074\000\046\000\047\000\048\000\075\000\120\000\114\000\
\029\000\170\000\049\000\030\000\050\000\165\000\121\000\144\000\
\145\000\146\000\147\000\076\000\003\000\049\000\177\000\050\000\
\130\000\007\000\003\000\050\000\166\000\131\000\132\000\133\000\
\134\000\135\000\136\000\128\000\137\000\139\000\140\000\038\000\
\039\000\040\000\041\000\042\000\043\000\031\000\155\000\013\000\
\044\000\104\000\045\000\015\000\156\000\105\000\106\000\046\000\
\047\000\048\000\141\000\142\000\143\000\038\000\039\000\040\000\
\068\000\042\000\043\000\164\000\016\000\064\000\044\000\065\000\
\008\000\022\000\049\000\022\000\050\000\046\000\047\000\048\000\
\068\000\042\000\043\000\086\000\020\000\174\000\069\000\068\000\
\042\000\043\000\024\000\022\000\024\000\069\000\168\000\023\000\
\049\000\047\000\050\000\047\000\024\000\047\000\087\000\047\000\
\175\000\099\000\100\000\101\000\102\000\103\000\088\000\047\000\
\049\000\089\000\050\000\037\000\027\000\028\000\036\000\049\000\
\082\000\050\000\082\000\082\000\082\000\107\000\082\000\082\000\
\047\000\108\000\029\000\082\000\082\000\109\000\082\000\082\000\
\082\000\082\000\066\000\071\000\082\000\082\000\082\000\082\000\
\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
\082\000\075\000\083\000\075\000\072\000\075\000\045\000\075\000\
\045\000\073\000\045\000\085\000\045\000\075\000\097\000\075\000\
\075\000\075\000\075\000\098\000\045\000\075\000\075\000\075\000\
\075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
\075\000\075\000\080\000\110\000\080\000\112\000\080\000\036\000\
\080\000\036\000\113\000\036\000\115\000\116\000\080\000\123\000\
\080\000\080\000\080\000\080\000\117\000\036\000\080\000\080\000\
\080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
\080\000\080\000\080\000\068\000\118\000\068\000\122\000\068\000\
\126\000\068\000\091\000\092\000\093\000\094\000\095\000\096\000\
\124\000\068\000\068\000\127\000\016\000\148\000\149\000\068\000\
\068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
\068\000\068\000\068\000\068\000\065\000\150\000\065\000\151\000\
\065\000\125\000\065\000\152\000\159\000\153\000\154\000\158\000\
\169\000\172\000\065\000\065\000\155\000\163\000\167\000\173\000\
\065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
\065\000\065\000\065\000\065\000\065\000\066\000\156\000\066\000\
\171\000\066\000\176\000\066\000\178\000\009\000\005\000\010\000\
\006\000\085\000\024\000\066\000\066\000\010\000\019\000\026\000\
\111\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
\066\000\066\000\066\000\066\000\066\000\066\000\067\000\051\000\
\067\000\162\000\067\000\138\000\067\000\082\000\084\000\000\000\
\000\000\000\000\000\000\000\000\067\000\067\000\000\000\000\000\
\000\000\000\000\067\000\067\000\067\000\067\000\067\000\067\000\
\067\000\067\000\067\000\067\000\067\000\067\000\067\000\064\000\
\000\000\064\000\000\000\064\000\000\000\064\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\064\000\064\000\000\000\
\000\000\000\000\000\000\064\000\064\000\064\000\064\000\064\000\
\064\000\000\000\000\000\000\000\000\000\000\000\064\000\064\000\
\032\000\032\000\000\000\032\000\000\000\032\000\000\000\000\000\
\000\000\000\000\000\000\000\000\000\000\000\000\000\000\032\000\
\000\000\000\000\000\000\032\000\082\000\000\000\000\000\000\000\
\082\000\000\000\000\000\032\000\082\000\082\000\032\000\125\000\
\000\000\082\000\082\000\000\000\000\000\082\000\082\000\082\000\
\082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
\000\000\091\000\092\000\093\000\094\000\095\000\096\000"

let yycheck = "\037\000\
\000\000\045\000\004\001\005\001\006\001\031\000\044\000\044\000\
\010\001\035\000\001\001\002\001\003\001\004\001\005\001\006\001\
\005\001\004\001\005\001\010\001\011\001\012\001\024\001\010\001\
\001\000\010\001\017\001\018\001\019\001\014\001\074\000\069\000\
\005\001\158\000\036\001\008\001\038\001\009\001\075\000\104\000\
\105\000\106\000\107\000\028\001\039\001\036\001\171\000\038\001\
\086\000\004\001\039\001\038\001\024\001\091\000\092\000\093\000\
\094\000\095\000\096\000\085\000\097\000\099\000\100\000\001\001\
\002\001\003\001\004\001\005\001\006\001\042\001\014\001\022\001\
\010\001\021\001\012\001\022\001\020\001\025\001\026\001\017\001\
\018\001\019\001\101\000\102\000\103\000\001\001\002\001\003\001\
\004\001\005\001\006\001\129\000\004\001\023\001\010\001\025\001\
\005\001\011\001\036\001\013\001\038\001\017\001\018\001\019\001\
\004\001\005\001\006\001\008\001\004\001\009\001\010\001\004\001\
\005\001\006\001\011\001\023\001\013\001\010\001\155\000\004\001\
\036\001\009\001\038\001\011\001\022\001\013\001\027\001\015\001\
\166\000\035\001\036\001\037\001\038\001\039\001\035\001\023\001\
\036\001\038\001\038\001\007\001\024\001\023\001\004\001\036\001\
\009\001\038\001\011\001\012\001\013\001\012\001\015\001\016\001\
\040\001\016\001\005\001\020\001\021\001\020\001\023\001\024\001\
\025\001\026\001\016\001\010\001\029\001\030\001\031\001\032\001\
\033\001\034\001\035\001\036\001\037\001\038\001\039\001\040\001\
\041\001\009\001\006\001\011\001\010\001\013\001\009\001\015\001\
\011\001\010\001\013\001\009\001\015\001\021\001\040\001\023\001\
\024\001\025\001\026\001\041\001\023\001\029\001\030\001\031\001\
\032\001\033\001\034\001\035\001\036\001\037\001\038\001\039\001\
\040\001\041\001\009\001\016\001\011\001\004\001\013\001\009\001\
\015\001\011\001\010\001\013\001\023\001\004\001\021\001\010\001\
\023\001\024\001\025\001\026\001\004\001\023\001\029\001\030\001\
\031\001\032\001\033\001\034\001\035\001\036\001\037\001\038\001\
\039\001\040\001\041\001\009\001\004\001\011\001\004\001\013\001\
\023\001\015\001\029\001\030\001\031\001\032\001\033\001\034\001\
\011\001\023\001\024\001\013\001\004\001\004\001\004\001\029\001\
\030\001\031\001\032\001\033\001\034\001\035\001\036\001\037\001\
\038\001\039\001\040\001\041\001\009\001\004\001\011\001\011\001\
\013\001\011\001\015\001\011\001\015\001\011\001\011\001\011\001\
\004\001\004\001\023\001\024\001\014\001\016\001\013\001\009\001\
\029\001\030\001\031\001\032\001\033\001\034\001\035\001\036\001\
\037\001\038\001\039\001\040\001\041\001\009\001\020\001\011\001\
\011\001\013\001\015\001\015\001\009\001\004\001\023\001\004\001\
\023\001\016\001\011\001\023\001\024\001\005\000\011\000\017\000\
\065\000\029\001\030\001\031\001\032\001\033\001\034\001\035\001\
\036\001\037\001\038\001\039\001\040\001\041\001\009\001\030\000\
\011\001\126\000\013\001\098\000\015\001\048\000\050\000\255\255\
\255\255\255\255\255\255\255\255\023\001\024\001\255\255\255\255\
\255\255\255\255\029\001\030\001\031\001\032\001\033\001\034\001\
\035\001\036\001\037\001\038\001\039\001\040\001\041\001\009\001\
\255\255\011\001\255\255\013\001\255\255\015\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\023\001\024\001\255\255\
\255\255\255\255\255\255\029\001\030\001\031\001\032\001\033\001\
\034\001\255\255\255\255\255\255\255\255\255\255\040\001\041\001\
\008\001\009\001\255\255\011\001\255\255\013\001\255\255\255\255\
\255\255\255\255\255\255\255\255\255\255\255\255\255\255\023\001\
\255\255\255\255\255\255\027\001\012\001\255\255\255\255\255\255\
\016\001\255\255\255\255\035\001\020\001\021\001\038\001\011\001\
\255\255\025\001\026\001\255\255\255\255\029\001\030\001\031\001\
\032\001\033\001\034\001\035\001\036\001\037\001\038\001\039\001\
\255\255\029\001\030\001\031\001\032\001\033\001\034\001"

let yynames_const = "\
  CALL_OF\000\
  RETURN_OF\000\
  CALLORRETURN_OF\000\
  AFF\000\
  LCURLY\000\
  RCURLY\000\
  LPAREN\000\
  RPAREN\000\
  LSQUARE\000\
  RSQUARE\000\
  LBRACELBRACE\000\
  RBRACERBRACE\000\
  RARROW\000\
  TRUE\000\
  FALSE\000\
  NOT\000\
  DOT\000\
  AMP\000\
  COLON\000\
  SEMI_COLON\000\
  COMMA\000\
  PIPE\000\
  CARET\000\
  QUESTION\000\
  COLUMNCOLUMN\000\
  EQ\000\
  LT\000\
  GT\000\
  LE\000\
  GE\000\
  NEQ\000\
  PLUS\000\
  MINUS\000\
  SLASH\000\
  STAR\000\
  PERCENT\000\
  OR\000\
  AND\000\
  OTHERWISE\000\
  EOF\000\
  "

let yynames_block = "\
  IDENTIFIER\000\
  METAVAR\000\
  INT\000\
  "

let yyact = [|
  (fun _ -> failwith "parser")
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : 'options) in
    let _2 = (Parsing.peek_val __caml_parser_env 2 : 'metavars) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'states) in
    Obj.repr(
# 175 "src/plugins/aorai/yaparser.mly"
                                   ( build_automaton _1 _2 _3 )
# 513 "src/plugins/aorai/yaparser.ml"
               : Promelaast.parsed_automaton))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'options) in
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'option) in
    Obj.repr(
# 178 "src/plugins/aorai/yaparser.mly"
                   ( _1 @ [_2] )
# 521 "src/plugins/aorai/yaparser.ml"
               : 'options))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'option) in
    Obj.repr(
# 179 "src/plugins/aorai/yaparser.mly"
                   ( [_1] )
# 528 "src/plugins/aorai/yaparser.ml"
               : 'options))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 2 : string) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'opt_identifiers) in
    Obj.repr(
# 183 "src/plugins/aorai/yaparser.mly"
                                                  ( 
    match _2 with
    | "init" -> Init _3
    | "accept" -> Accept _3
    | "deterministic" -> Deterministic
    | "observables" -> Observables _3
    | _ ->  Aorai_option.abort "unknown option: '%s'" _2
  )
# 543 "src/plugins/aorai/yaparser.ml"
               : 'option))
; (fun __caml_parser_env ->
    Obj.repr(
# 193 "src/plugins/aorai/yaparser.mly"
                ( [] )
# 549 "src/plugins/aorai/yaparser.ml"
               : 'opt_identifiers))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'id_list) in
    Obj.repr(
# 194 "src/plugins/aorai/yaparser.mly"
                  ( _2 )
# 556 "src/plugins/aorai/yaparser.ml"
               : 'opt_identifiers))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'id_list) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 198 "src/plugins/aorai/yaparser.mly"
                             ( _1 @ [_3] )
# 564 "src/plugins/aorai/yaparser.ml"
               : 'id_list))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 199 "src/plugins/aorai/yaparser.mly"
                             ( [_1] )
# 571 "src/plugins/aorai/yaparser.ml"
               : 'id_list))
; (fun __caml_parser_env ->
    Obj.repr(
# 203 "src/plugins/aorai/yaparser.mly"
                       ( [] )
# 577 "src/plugins/aorai/yaparser.ml"
               : 'metavars))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'non_empty_metavars) in
    Obj.repr(
# 204 "src/plugins/aorai/yaparser.mly"
                       ( _1 )
# 584 "src/plugins/aorai/yaparser.ml"
               : 'metavars))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'non_empty_metavars) in
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'metavar) in
    Obj.repr(
# 207 "src/plugins/aorai/yaparser.mly"
                               ( _1 @ [_2] )
# 592 "src/plugins/aorai/yaparser.ml"
               : 'non_empty_metavars))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'metavar) in
    Obj.repr(
# 208 "src/plugins/aorai/yaparser.mly"
                               ( [_1] )
# 599 "src/plugins/aorai/yaparser.ml"
               : 'non_empty_metavars))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : string) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : string) in
    Obj.repr(
# 212 "src/plugins/aorai/yaparser.mly"
                                        ( _1,_3 )
# 607 "src/plugins/aorai/yaparser.ml"
               : 'metavar))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'states) in
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'state) in
    Obj.repr(
# 216 "src/plugins/aorai/yaparser.mly"
                 ( _1@_2 )
# 615 "src/plugins/aorai/yaparser.ml"
               : 'states))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'state) in
    Obj.repr(
# 217 "src/plugins/aorai/yaparser.mly"
          ( _1 )
# 622 "src/plugins/aorai/yaparser.ml"
               : 'states))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : string) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'transitions) in
    Obj.repr(
# 221 "src/plugins/aorai/yaparser.mly"
                                            (
      let start_state = fetch_and_create_state _1 in
      let (_, transitions) =
        List.fold_left
          (fun (otherwise, transitions) (cross,actions,stop_state) ->
            if otherwise then
              Aorai_option.abort
                "'other' directive in definition of %s \
                transitions is not the last one" start_state.name
            else begin
              let trans =
                { start=start_state; stop=stop_state;
                  cross; actions; numt=(-1) }::transitions
              in
              let otherwise =
                match cross with
                  | Otherwise -> true
                  | Seq _ -> false
              in otherwise, trans
            end)
          (false,[]) _3
      in
      List.rev transitions
  )
# 653 "src/plugins/aorai/yaparser.ml"
               : 'state))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'transitions) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'transition) in
    Obj.repr(
# 247 "src/plugins/aorai/yaparser.mly"
                                ( _1@[_3] )
# 661 "src/plugins/aorai/yaparser.ml"
               : 'transitions))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'transition) in
    Obj.repr(
# 248 "src/plugins/aorai/yaparser.mly"
               ( [_1] )
# 668 "src/plugins/aorai/yaparser.ml"
               : 'transitions))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 4 : 'seq_elt) in
    let _4 = (Parsing.peek_val __caml_parser_env 2 : 'actions) in
    let _6 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 254 "src/plugins/aorai/yaparser.mly"
      ( (Seq _2, _4, prefetch_and_create_state _6) )
# 677 "src/plugins/aorai/yaparser.ml"
               : 'transition))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 2 : 'actions) in
    let _4 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 256 "src/plugins/aorai/yaparser.mly"
      ( (Otherwise, _2, prefetch_and_create_state _4) )
# 685 "src/plugins/aorai/yaparser.ml"
               : 'transition))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'actions) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 258 "src/plugins/aorai/yaparser.mly"
      ( (Seq (to_seq PTrue), _1, prefetch_and_create_state _3) )
# 693 "src/plugins/aorai/yaparser.ml"
               : 'transition))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'seq_elt) in
    Obj.repr(
# 262 "src/plugins/aorai/yaparser.mly"
            ( _1 )
# 700 "src/plugins/aorai/yaparser.ml"
               : 'non_empty_seq))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'seq_elt) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'seq) in
    Obj.repr(
# 263 "src/plugins/aorai/yaparser.mly"
                           ( _1 @ _3 )
# 708 "src/plugins/aorai/yaparser.ml"
               : 'non_empty_seq))
; (fun __caml_parser_env ->
    Obj.repr(
# 267 "src/plugins/aorai/yaparser.mly"
                  ( [] )
# 714 "src/plugins/aorai/yaparser.ml"
               : 'seq))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'non_empty_seq) in
    Obj.repr(
# 268 "src/plugins/aorai/yaparser.mly"
                  ( _1 )
# 721 "src/plugins/aorai/yaparser.ml"
               : 'seq))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'non_empty_seq) in
    Obj.repr(
# 272 "src/plugins/aorai/yaparser.mly"
                                  ( _2 )
# 728 "src/plugins/aorai/yaparser.ml"
               : 'guard))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 5 : string) in
    let _2 = (Parsing.peek_val __caml_parser_env 4 : 'pre_cond) in
    let _4 = (Parsing.peek_val __caml_parser_env 2 : 'seq) in
    let _6 = (Parsing.peek_val __caml_parser_env 0 : 'post_cond) in
    Obj.repr(
# 274 "src/plugins/aorai/yaparser.mly"
      ( let pre_cond =
          match _2 with
            | Behavior b -> PCall(_1,Some b)
            | Pre c -> PAnd (PCall(_1,None), c)
        in
        let post_cond =
          match _6 with
            | None -> PReturn _1
            | Some c -> PAnd (PReturn _1,c)
        in
        (to_seq pre_cond) @ _4 @ to_seq post_cond
      )
# 749 "src/plugins/aorai/yaparser.ml"
               : 'guard))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 4 : string) in
    let _3 = (Parsing.peek_val __caml_parser_env 2 : 'non_empty_seq) in
    let _5 = (Parsing.peek_val __caml_parser_env 0 : 'post_cond) in
    Obj.repr(
# 287 "src/plugins/aorai/yaparser.mly"
      ( let post_cond =
          match _5 with
            | None -> PReturn _1
            | Some c -> PAnd (PReturn _1,c)
        in
        (to_seq (PCall (_1, None))) @ _3 @ to_seq post_cond
      )
# 764 "src/plugins/aorai/yaparser.ml"
               : 'guard))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : string) in
    let _4 = (Parsing.peek_val __caml_parser_env 0 : 'post_cond) in
    Obj.repr(
# 295 "src/plugins/aorai/yaparser.mly"
      ( let post_cond =
          match _4 with
            | None -> PReturn _1
            | Some c -> PAnd (PReturn _1,c)
        in
        (to_seq (PCall (_1, None))) @ to_seq post_cond
      )
# 778 "src/plugins/aorai/yaparser.ml"
               : 'guard))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 305 "src/plugins/aorai/yaparser.mly"
                            ( Behavior _2 )
# 785 "src/plugins/aorai/yaparser.ml"
               : 'pre_cond))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'cond) in
    Obj.repr(
# 306 "src/plugins/aorai/yaparser.mly"
                                   ( Pre _2 )
# 792 "src/plugins/aorai/yaparser.ml"
               : 'pre_cond))
; (fun __caml_parser_env ->
    Obj.repr(
# 310 "src/plugins/aorai/yaparser.mly"
                  ( None )
# 798 "src/plugins/aorai/yaparser.ml"
               : 'post_cond))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'cond) in
    Obj.repr(
# 311 "src/plugins/aorai/yaparser.mly"
                                   ( Some _2 )
# 805 "src/plugins/aorai/yaparser.ml"
               : 'post_cond))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'cond) in
    Obj.repr(
# 315 "src/plugins/aorai/yaparser.mly"
         ( to_seq _1 )
# 812 "src/plugins/aorai/yaparser.ml"
               : 'seq_elt))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'guard) in
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'repetition) in
    Obj.repr(
# 316 "src/plugins/aorai/yaparser.mly"
                     (
    let min, max = _2 in
    match _1 with
      | [ s ] when Data_for_aorai.is_single s ->
        [ { s with min_rep = min; max_rep = max } ]
      | l ->
        if is_no_repet (min,max) then
          l (* [ a; [b;c]; d] is equivalent to [a;b;c;d] *)
        else [ { condition = None; nested = l; min_rep = min; max_rep = max } ]
  )
# 829 "src/plugins/aorai/yaparser.ml"
               : 'seq_elt))
; (fun __caml_parser_env ->
    Obj.repr(
# 330 "src/plugins/aorai/yaparser.mly"
      ( Some Data_for_aorai.cst_one, Some Data_for_aorai.cst_one )
# 835 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    Obj.repr(
# 331 "src/plugins/aorai/yaparser.mly"
         ( Some Data_for_aorai.cst_one, None)
# 841 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    Obj.repr(
# 332 "src/plugins/aorai/yaparser.mly"
         ( None, None )
# 847 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    Obj.repr(
# 333 "src/plugins/aorai/yaparser.mly"
             ( None, Some Data_for_aorai.cst_one )
# 853 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 3 : 'arith_relation) in
    let _4 = (Parsing.peek_val __caml_parser_env 1 : 'arith_relation) in
    Obj.repr(
# 334 "src/plugins/aorai/yaparser.mly"
                                                      ( Some _2, Some _4 )
# 861 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'arith_relation) in
    Obj.repr(
# 335 "src/plugins/aorai/yaparser.mly"
                                 ( Some _2, Some _2 )
# 868 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation) in
    Obj.repr(
# 336 "src/plugins/aorai/yaparser.mly"
                                       ( Some _2, None )
# 875 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'arith_relation) in
    Obj.repr(
# 337 "src/plugins/aorai/yaparser.mly"
                                       ( None, Some _3 )
# 882 "src/plugins/aorai/yaparser.ml"
               : 'repetition))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'conjunction) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'cond) in
    Obj.repr(
# 340 "src/plugins/aorai/yaparser.mly"
                        ( POr (_1,_3) )
# 890 "src/plugins/aorai/yaparser.ml"
               : 'cond))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'conjunction) in
    Obj.repr(
# 341 "src/plugins/aorai/yaparser.mly"
                ( _1 )
# 897 "src/plugins/aorai/yaparser.ml"
               : 'cond))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'atomic_cond) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'conjunction) in
    Obj.repr(
# 344 "src/plugins/aorai/yaparser.mly"
                                ( PAnd(_1,_3) )
# 905 "src/plugins/aorai/yaparser.ml"
               : 'conjunction))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_cond) in
    Obj.repr(
# 345 "src/plugins/aorai/yaparser.mly"
                ( _1 )
# 912 "src/plugins/aorai/yaparser.ml"
               : 'conjunction))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 1 : string) in
    Obj.repr(
# 349 "src/plugins/aorai/yaparser.mly"
      ( POr (PCall (_3,None), PReturn _3) )
# 919 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 1 : string) in
    Obj.repr(
# 350 "src/plugins/aorai/yaparser.mly"
                                      ( PCall (_3,None) )
# 926 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    let _3 = (Parsing.peek_val __caml_parser_env 1 : string) in
    Obj.repr(
# 351 "src/plugins/aorai/yaparser.mly"
                                        ( PReturn _3 )
# 933 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    Obj.repr(
# 352 "src/plugins/aorai/yaparser.mly"
         ( PTrue )
# 939 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    Obj.repr(
# 353 "src/plugins/aorai/yaparser.mly"
          ( PFalse )
# 945 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'atomic_cond) in
    Obj.repr(
# 354 "src/plugins/aorai/yaparser.mly"
                    ( PNot _2 )
# 952 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'cond) in
    Obj.repr(
# 355 "src/plugins/aorai/yaparser.mly"
                       ( _2 )
# 959 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'logic_relation) in
    Obj.repr(
# 356 "src/plugins/aorai/yaparser.mly"
                   ( _1 )
# 966 "src/plugins/aorai/yaparser.ml"
               : 'atomic_cond))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 360 "src/plugins/aorai/yaparser.mly"
                                     ( PRel(Eq, _1, _3) )
# 974 "src/plugins/aorai/yaparser.ml"
               : 'logic_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 361 "src/plugins/aorai/yaparser.mly"
                                     ( PRel(Lt, _1, _3) )
# 982 "src/plugins/aorai/yaparser.ml"
               : 'logic_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 362 "src/plugins/aorai/yaparser.mly"
                                     ( PRel(Gt, _1, _3) )
# 990 "src/plugins/aorai/yaparser.ml"
               : 'logic_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 363 "src/plugins/aorai/yaparser.mly"
                                     ( PRel(Le, _1, _3) )
# 998 "src/plugins/aorai/yaparser.ml"
               : 'logic_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 364 "src/plugins/aorai/yaparser.mly"
                                     ( PRel(Ge, _1, _3) )
# 1006 "src/plugins/aorai/yaparser.ml"
               : 'logic_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 365 "src/plugins/aorai/yaparser.mly"
                                      ( PRel(Neq, _1, _3) )
# 1014 "src/plugins/aorai/yaparser.ml"
               : 'logic_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_mul) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 370 "src/plugins/aorai/yaparser.mly"
                                           ( PBinop(Badd,_1,_3) )
# 1022 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_mul) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation) in
    Obj.repr(
# 371 "src/plugins/aorai/yaparser.mly"
                                            ( PBinop(Bsub,_1,_3) )
# 1030 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation_mul) in
    Obj.repr(
# 372 "src/plugins/aorai/yaparser.mly"
                       ( _1 )
# 1037 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_mul) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation_bw) in
    Obj.repr(
# 376 "src/plugins/aorai/yaparser.mly"
                                               ( PBinop(Bdiv,_1,_3) )
# 1045 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_mul))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_mul) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation_bw) in
    Obj.repr(
# 377 "src/plugins/aorai/yaparser.mly"
                                              ( PBinop(Bmul, _1, _3) )
# 1053 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_mul))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_mul) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation_bw) in
    Obj.repr(
# 378 "src/plugins/aorai/yaparser.mly"
                                                 ( PBinop(Bmod, _1, _3) )
# 1061 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_mul))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'arith_relation_bw) in
    Obj.repr(
# 379 "src/plugins/aorai/yaparser.mly"
                      ( _1 )
# 1068 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_mul))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'access_or_const) in
    Obj.repr(
# 383 "src/plugins/aorai/yaparser.mly"
                    ( _1 )
# 1075 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_bw))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_bw) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'access_or_const) in
    Obj.repr(
# 384 "src/plugins/aorai/yaparser.mly"
                                          ( PBinop(Bbw_and,_1,_3) )
# 1083 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_bw))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_bw) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'access_or_const) in
    Obj.repr(
# 385 "src/plugins/aorai/yaparser.mly"
                                           ( PBinop(Bbw_or,_1,_3) )
# 1091 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_bw))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'arith_relation_bw) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : 'access_or_const) in
    Obj.repr(
# 386 "src/plugins/aorai/yaparser.mly"
                                            ( PBinop(Bbw_xor,_1,_3) )
# 1099 "src/plugins/aorai/yaparser.ml"
               : 'arith_relation_bw))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 390 "src/plugins/aorai/yaparser.mly"
        ( PCst (IntConstant _1) )
# 1106 "src/plugins/aorai/yaparser.ml"
               : 'access_or_const))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 391 "src/plugins/aorai/yaparser.mly"
              ( PUnop (Uminus, PCst (IntConstant _2)) )
# 1113 "src/plugins/aorai/yaparser.ml"
               : 'access_or_const))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'access) in
    Obj.repr(
# 392 "src/plugins/aorai/yaparser.mly"
           ( _1 )
# 1120 "src/plugins/aorai/yaparser.ml"
               : 'access_or_const))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'access) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 397 "src/plugins/aorai/yaparser.mly"
                          ( PField(_1,_3) )
# 1128 "src/plugins/aorai/yaparser.ml"
               : 'access))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 2 : 'access) in
    let _3 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 398 "src/plugins/aorai/yaparser.mly"
                             ( PField(PUnop(Ustar,_1),_3) )
# 1136 "src/plugins/aorai/yaparser.ml"
               : 'access))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : 'access) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'access_or_const) in
    Obj.repr(
# 399 "src/plugins/aorai/yaparser.mly"
                                           ( PArrget(_1,_3) )
# 1144 "src/plugins/aorai/yaparser.ml"
               : 'access))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : 'access_leaf) in
    Obj.repr(
# 400 "src/plugins/aorai/yaparser.mly"
                    (_1)
# 1151 "src/plugins/aorai/yaparser.ml"
               : 'access))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'access) in
    Obj.repr(
# 404 "src/plugins/aorai/yaparser.mly"
                ( PUnop (Ustar,_2) )
# 1158 "src/plugins/aorai/yaparser.ml"
               : 'access_leaf))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 4 : string) in
    let _5 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 405 "src/plugins/aorai/yaparser.mly"
                                            ( PPrm(_1,_5) )
# 1166 "src/plugins/aorai/yaparser.ml"
               : 'access_leaf))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 406 "src/plugins/aorai/yaparser.mly"
               ( PVar _1 )
# 1173 "src/plugins/aorai/yaparser.ml"
               : 'access_leaf))
; (fun __caml_parser_env ->
    let _2 = (Parsing.peek_val __caml_parser_env 1 : 'arith_relation) in
    Obj.repr(
# 407 "src/plugins/aorai/yaparser.mly"
                                 ( _2 )
# 1180 "src/plugins/aorai/yaparser.ml"
               : 'access_leaf))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 0 : string) in
    Obj.repr(
# 408 "src/plugins/aorai/yaparser.mly"
            ( PMetavar _1 )
# 1187 "src/plugins/aorai/yaparser.ml"
               : 'access_leaf))
; (fun __caml_parser_env ->
    Obj.repr(
# 412 "src/plugins/aorai/yaparser.mly"
                                    ( [] )
# 1193 "src/plugins/aorai/yaparser.ml"
               : 'actions))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 1 : 'action) in
    let _2 = (Parsing.peek_val __caml_parser_env 0 : 'actions) in
    Obj.repr(
# 413 "src/plugins/aorai/yaparser.mly"
                   ( _1 :: _2 )
# 1201 "src/plugins/aorai/yaparser.ml"
               : 'actions))
; (fun __caml_parser_env ->
    let _1 = (Parsing.peek_val __caml_parser_env 3 : string) in
    let _3 = (Parsing.peek_val __caml_parser_env 1 : 'arith_relation) in
    Obj.repr(
# 417 "src/plugins/aorai/yaparser.mly"
                                          ( Metavar_assign (_1, _3) )
# 1209 "src/plugins/aorai/yaparser.ml"
               : 'action))
(* Entry main *)
; (fun __caml_parser_env -> raise (Parsing.YYexit (Parsing.peek_val __caml_parser_env 0)))
|]
let yytables =
  { Parsing.actions=yyact;
    Parsing.transl_const=yytransl_const;
    Parsing.transl_block=yytransl_block;
    Parsing.lhs=yylhs;
    Parsing.len=yylen;
    Parsing.defred=yydefred;
    Parsing.dgoto=yydgoto;
    Parsing.sindex=yysindex;
    Parsing.rindex=yyrindex;
    Parsing.gindex=yygindex;
    Parsing.tablesize=yytablesize;
    Parsing.table=yytable;
    Parsing.check=yycheck;
    Parsing.error_function=parse_error;
    Parsing.names_const=yynames_const;
    Parsing.names_block=yynames_block }
let main (lexfun : Lexing.lexbuf -> token) (lexbuf : Lexing.lexbuf) =
   (Parsing.yyparse yytables 1 lexfun lexbuf : Promelaast.parsed_automaton)
